/*****************************************************************************
				Tejas Simulator
------------------------------------------------------------------------------------------------------------

   Copyright [2010] [Indian Institute of Technology, Delhi]
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
------------------------------------------------------------------------------------------------------------

	Contributors:  Moksh Upadhyay, Abhishek Sagar
*****************************************************************************/
package config;

import emulatorinterface.communication.IpcBase;
import generic.MultiPortingType;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import net.*;
import net.NOC.CONNECTIONTYPE;
import memorysystem.Cache;
import memorysystem.Cache.CacheType;
import memorysystem.Cache.CoherenceType;

import org.w3c.dom.*;

import config.BranchPredictorConfig.*;

import memorysystem.nuca.NucaCache.Mapping;
import memorysystem.nuca.NucaCache.NucaType;

import generic.PortType;

//<Cache name="iCache" nextLevel="L2" nextLevelId="$i/4" firstLevel="true" type="ICache_32K_4"/>
//<Cache name="l1Cache" nextLevel="L2" nextLevelId="$i/4" firstLevel="true" type="L1Cache_32K_4"/>
//<Cache name="L2" numComponents="2" nextLevel="L3" type="L2Cache_1M_8"/>

public class XMLParser 
{
	private static Document doc;

	public static void parse(String fileName) 
	{ 
		try 
		{
			File file = new File(fileName);
			DocumentBuilderFactory DBFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder DBuilder = DBFactory.newDocumentBuilder();
			doc = DBuilder.parse(file);
			doc.getDocumentElement().normalize();
			//System.out.println("Root element : " + doc.getDocumentElement().getNodeName());
			
			createSharedCacheConfigs();
			setEmulatorParameters();
			setSimulationParameters();
			
			setSystemParameters();
		} 
		catch (Exception e) 
		{
			// Cannot use shutdown handler here 
			// the stats file is not setup yet
			// so printing stats in the shutdown hook will halt the display of error message 
			// so the exact error cannot be debugged
			//System.err.println("Error in reading config file : " + e);
			//System.err.flush();
			//new Exception().printStackTrace();
			//System.exit(0);
			misc.Error.showErrorAndExit("Error in reading config file : " + e);
		}
 	}
	
	private static void createSharedCacheConfigs() throws Exception {
		Element sharedCachesNode = (Element)doc.getElementsByTagName("SharedCaches").item(0);
		
		NodeList nodeLst = sharedCachesNode.getElementsByTagName("Cache");
		if (nodeLst.item(0) == null) {
			System.out.println("Shared caches not found !!");
		}
		
		for(int i=0; i<nodeLst.getLength(); i++) {
			Element cacheNode = (Element)nodeLst.item(i);			
			CacheConfig config = createCacheConfig(cacheNode);
			SystemConfig.declaredCacheConfigs.add(config);
		}
	}
	
	private static CacheConfig createCacheConfig(Element cacheNode) {
		CacheConfig config = new CacheConfig();
		
		config.cacheName = cacheNode.getAttribute("name");
		
		if(isAttributePresent(cacheNode, "firstLevel")) {
			config.firstLevel = 
				Boolean.parseBoolean(cacheNode.getAttribute("firstLevel"));
		} else {
			config.firstLevel = false;
		}
		
		if(isAttributePresent(cacheNode, "numComponents")) {
			config.numComponents = 
				Integer.parseInt(cacheNode.getAttribute("numComponents"));
		} else {
			config.numComponents = 1;
		}
		
		config.nextLevel=cacheNode.getAttribute("nextLevel");
		if(isAttributePresent(cacheNode, "nextLevelId")) {
			config.nextLevelId = cacheNode.getAttribute("nextLevelId"); 
		}
	
		String cacheType = cacheNode.getAttribute("type");
		
		Element cacheTypeElmnt = searchLibraryForItem(cacheType);
		setCacheProperties(cacheTypeElmnt, config);
		
		return config;
	}
	
	private static boolean isAttributePresent(Element element, String str) {
		return (element.getAttribute(str)!="");
	}
	
	private static boolean isElementPresent(String tagName, Element parent) // Get the immediate string value of a particular tag name under a particular parent tag
	{
		NodeList nodeLst = parent.getElementsByTagName(tagName);
		if (nodeLst.item(0) == null) {
			return false;
		} else {
			return true;
		}
	}
	
	// For an ith core specified, mark the ith field of this long as 1. 
	private static long parseMapper (String s) 
	{
		long ret = 0;
		String delims = "[,]+";
		String[] tokens = s.split(delims);
		for (int i=0; i<tokens.length; i++) 
		{
			String delimsin = "[-]+";
			String[] tokensin = tokens[i].split(delimsin);
			if (tokensin.length == 1) 
			{
				ret = ret | (1 << Integer.parseInt(tokensin[0]));
			}
			else if (tokensin.length == 2) 
			{
				int start = Integer.parseInt(tokensin[0]);
				int end = Integer.parseInt(tokensin[1]);
				for (int j=start; j<=end; j++) 
				{
					ret = ret | (1 << j);
				}
			}
			else 
			{
				System.out.println("Physical Core mapping not correct in config.xml");
				System.exit(0);
			}
		}
		return ret;
	}
		
		
	private static void setEmulatorParameters() {
		NodeList nodeLst = doc.getElementsByTagName("Emulator");
		Node emulatorNode = nodeLst.item(0);
		Element emulatorElmnt = (Element) emulatorNode;
		
		EmulatorConfig.EmulatorType = Integer.parseInt(getImmediateString("EmulatorType", emulatorElmnt));
		EmulatorConfig.CommunicationType = Integer.parseInt(getImmediateString("CommunicationType", emulatorElmnt));
		EmulatorConfig.PinTool = getImmediateString("PinTool", emulatorElmnt);
		EmulatorConfig.PinInstrumentor = getImmediateString("PinInstrumentor", emulatorElmnt);
		EmulatorConfig.QemuTool = getImmediateString("QemuTool", emulatorElmnt);
		EmulatorConfig.ShmLibDirectory = getImmediateString("ShmLibDirectory", emulatorElmnt);
		EmulatorConfig.KillEmulatorScript = getImmediateString("KillEmulatorScript", emulatorElmnt);
	}
	
	private static void setSimulationParameters()
	{
		NodeList nodeLst = doc.getElementsByTagName("Simulation");
		Node simulationNode = nodeLst.item(0);
		Element simulationElmnt = (Element) simulationNode;
		SimulationConfig.NumTempIntReg = Integer.parseInt(getImmediateString("NumTempIntReg", simulationElmnt));
		SimulationConfig.NumInsToIgnore = Long.parseLong(getImmediateString("NumInsToIgnore", simulationElmnt));
		
		SimulationConfig.collectInsnWorkingSetInfo = 
				Boolean.parseBoolean(getImmediateString("CollectInsnWorkingSet", simulationElmnt));
		
		SimulationConfig.insnWorkingSetChunkSize = 
				Long.parseLong(getImmediateString("InsnWorkingSetChunkSize", simulationElmnt));
		
		SimulationConfig.collectDataWorkingSetInfo = 
				Boolean.parseBoolean(getImmediateString("CollectDataWorkingSet", simulationElmnt));
		
		SimulationConfig.dataWorkingSetChunkSize = 
				Long.parseLong(getImmediateString("DataWorkingSetChunkSize", simulationElmnt));
		
		int tempVal = Integer.parseInt(getImmediateString("IndexAddrModeEnable", simulationElmnt));
		if (tempVal == 0)
			SimulationConfig.IndexAddrModeEnable = false;
		else
			SimulationConfig.IndexAddrModeEnable = true;
		
		SimulationConfig.MapEmuCores = parseMapper(getImmediateString("EmuCores", simulationElmnt));
		SimulationConfig.MapJavaCores = parseMapper(getImmediateString("JavaCores", simulationElmnt));
		
		//System.out.println(SimulationConfig.NumTempIntReg + ", " + SimulationConfig.IndexAddrModeEnable);
		
		if(getImmediateString("DebugMode", simulationElmnt).compareTo("true") == 0 ||
				getImmediateString("DebugMode", simulationElmnt).compareTo("True") == 0)
		{
			SimulationConfig.debugMode = true;
		}
		else
		{
			SimulationConfig.debugMode = false;
		}
		if(getImmediateString("DetachMemSys", simulationElmnt).compareTo("true") == 0 ||
				getImmediateString("DetachMemSys", simulationElmnt).compareTo("True") == 0)
		{
			SimulationConfig.detachMemSys = true;
		}
		else
		{
			SimulationConfig.detachMemSys = false;
		}
		
		if(getImmediateString("writeToFile", simulationElmnt).compareTo("true") == 0 ||
				getImmediateString("writeToFile", simulationElmnt).compareTo("True") == 0)
		{
			SimulationConfig.writeToFile = true;
		}
		else
		{
			SimulationConfig.writeToFile = false;
		}
		
		SimulationConfig.numInstructionsToBeWritten = Integer.parseInt(getImmediateString("numInstructionsToBeWritten", simulationElmnt));
		SimulationConfig.InstructionsFilename = getImmediateString("InstructionsFilename", simulationElmnt);
		
		if(getImmediateString("subsetSim", simulationElmnt).compareTo("true") == 0 ||
				getImmediateString("subsetSim", simulationElmnt).compareTo("True") == 0)
		{
			SimulationConfig.subsetSimulation = true;
			SimulationConfig.subsetSimSize = Long.parseLong(getImmediateString("subsetSimSize", simulationElmnt));
		}
		else
		{
			SimulationConfig.subsetSimulation = false;
			SimulationConfig.subsetSimSize = -1;
		}
		
		if(getImmediateString("pinpointsSim", simulationElmnt).compareTo("true") == 0 ||
				getImmediateString("pinpointsSim", simulationElmnt).compareTo("True") == 0)
		{
			SimulationConfig.pinpointsSimulation = true;
			SimulationConfig.pinpointsFile = getImmediateString("pinpointsFile", simulationElmnt);
		}
		else
		{
			SimulationConfig.pinpointsSimulation = false;
			SimulationConfig.pinpointsFile = "";
		}
		
		if(getImmediateString("markerFunctions", simulationElmnt).compareTo("true") == 0 ||
				getImmediateString("markerFunctions", simulationElmnt).compareTo("True") == 0)
		{
			SimulationConfig.markerFunctionsSimulation = true;
			SimulationConfig.startMarker = getImmediateString("startSimMarker", simulationElmnt);
			SimulationConfig.endMarker = getImmediateString("endSimMarker", simulationElmnt);
		}
		else
		{
			SimulationConfig.markerFunctionsSimulation = false;
			SimulationConfig.startMarker = "";
			SimulationConfig.endMarker = "";
		}

		if(getImmediateString("PrintPowerStats", simulationElmnt).compareTo("true") == 0 ||
				getImmediateString("subsetSim", simulationElmnt).compareTo("True") == 0)
		{
			SimulationConfig.powerStats = true;
		}
		else
		{
			SimulationConfig.powerStats = false;
		}
		
		if(getImmediateString("Broadcast", simulationElmnt).toLowerCase().compareTo("true") == 0)
		{
			SimulationConfig.broadcast = true;
		}
		else
		{
			SimulationConfig.broadcast = false;
		}
		
		SimulationConfig.powerTrace = Integer.parseInt(getImmediateString("PowerTrace", simulationElmnt));
		SimulationConfig.numInsForTrace = Long.parseLong(getImmediateString("NumInsForTrace", simulationElmnt));
		SimulationConfig.numCyclesForTrace = Long.parseLong(getImmediateString("NumCyclesForTrace", simulationElmnt));

	}
	
	private static EnergyConfig getEnergyConfig(Element parent)
	{
		double leakageEnergy = Double.parseDouble(getImmediateString("LeakageEnergy", parent));
		double dynamicEnergy = Double.parseDouble(getImmediateString("DynamicEnergy", parent));
		
		EnergyConfig energyConfig = new EnergyConfig(leakageEnergy, dynamicEnergy);
		return energyConfig;
	}
	
	private static CacheEnergyConfig getCacheEnergyConfig(Element parent)
	{
		CacheEnergyConfig powerConfig = new CacheEnergyConfig();
		powerConfig.leakageEnergy = Double.parseDouble(getImmediateString("LeakageEnergy", parent));
		powerConfig.readDynamicEnergy = Double.parseDouble(getImmediateString("ReadDynamicEnergy", parent));
		powerConfig.writeDynamicEnergy = Double.parseDouble(getImmediateString("WriteDynamicEnergy", parent));
		return powerConfig;
	}
	
	private static void setSystemParameters()
	{
		NodeList nodeLst = doc.getElementsByTagName("System");
		Node systemNode = nodeLst.item(0);
		Element systemElmnt = (Element) systemNode;
		
		//Read number of cores and define the array of core configurations
		//Note that number of Cores specified in config.xml is deprecated and is instead done as follows
		//SystemConfig.maxNumJavaThreads = Integer.parseInt(getImmediateString("MaxNumJavaThreads", systemElmnt));
		SystemConfig.maxNumJavaThreads = 1;
		SystemConfig.numEmuThreadsPerJavaThread = Integer.parseInt(getImmediateString("NumEmuThreadsPerJavaThread", systemElmnt));
		
		SystemConfig.NoOfCores = SystemConfig.maxNumJavaThreads*SystemConfig.numEmuThreadsPerJavaThread;
						
		SystemConfig.mainMemoryLatency = Integer.parseInt(getImmediateString("MainMemoryLatency", systemElmnt));
		SystemConfig.mainMemoryFrequency = Long.parseLong(getImmediateString("MainMemoryFrequency", systemElmnt));
		SystemConfig.mainMemPortType = setPortType(getImmediateString("MainMemoryPortType", systemElmnt));
		SystemConfig.mainMemoryAccessPorts = Integer.parseInt(getImmediateString("MainMemoryAccessPorts", systemElmnt));
		SystemConfig.mainMemoryPortOccupancy = Integer.parseInt(getImmediateString("MainMemoryPortOccupancy", systemElmnt));
		
		Element mainMemElmnt = (Element)(systemElmnt.getElementsByTagName("MainMemory")).item(0);
		SystemConfig.mainMemoryControllerPower = getEnergyConfig(mainMemElmnt);
		
		Element globalClockElmnt = (Element)(systemElmnt.getElementsByTagName("GlobalClock")).item(0);
		SystemConfig.globalClockPower = getEnergyConfig(globalClockElmnt);
		
		SystemConfig.cacheBusLatency = Integer.parseInt(getImmediateString("CacheBusLatency", systemElmnt));

		SystemConfig.core = new CoreConfig[SystemConfig.NoOfCores];
		SystemConfig.directoryAccessLatency = Integer.parseInt(getImmediateString("directoryAccessLatency", systemElmnt));
		SystemConfig.memWBDelay = Integer.parseInt(getImmediateString("memWBDelay", systemElmnt));
		SystemConfig.dataTransferDelay = Integer.parseInt(getImmediateString("dataTransferDelay", systemElmnt));
		SystemConfig.invalidationSendDelay = Integer.parseInt(getImmediateString("invalidationSendDelay", systemElmnt));
		SystemConfig.invalidationAckCollectDelay = Integer.parseInt(getImmediateString("invalidationAckCollectDelay", systemElmnt));
		SystemConfig.ownershipChangeDelay = Integer.parseInt(getImmediateString("ownershipChangeDelay", systemElmnt));
		SystemConfig.dirNetworkDelay = Integer.parseInt(getImmediateString("dirNetworkDelay", systemElmnt));
	
		NodeList powerLst = doc.getElementsByTagName("Power");
		Node powerNode = powerLst.item(0);
		Element powerElmnt = (Element) powerNode;

		//Set core parameters
		NodeList coreLst = systemElmnt.getElementsByTagName("Core");
		//for (int i = 0; i < SystemConfig.NoOfCores; i++)
		for (int i = 0; i < SystemConfig.NoOfCores; i++)
		{
			SystemConfig.core[i] = new CoreConfig();
			CoreConfig core = SystemConfig.core[i]; //To be locally used for assignments
		
			Element coreElmnt = (Element) coreLst.item(0);
			
			core.frequency = Long.parseLong(getImmediateString("CoreFrequency", coreElmnt));
			
			core.pipelineType = PipelineType.valueOf(getImmediateString("PipelineType", coreElmnt));
			
			Element lsqElmnt = (Element)(coreElmnt.getElementsByTagName("LSQ")).item(0);
			core.LSQSize = Integer.parseInt(getImmediateString("LSQSize", lsqElmnt));
			core.LSQLatency = Integer.parseInt(getImmediateString("LSQLatency", lsqElmnt));
			core.LSQPortType = setPortType(getImmediateString("LSQPortType", lsqElmnt));
			core.LSQAccessPorts = Integer.parseInt(getImmediateString("LSQAccessPorts", lsqElmnt));
			core.LSQPortOccupancy = Integer.parseInt(getImmediateString("LSQPortOccupancy", lsqElmnt));
			core.LSQMultiportType = setMultiPortingType(getImmediateString("LSQMultiPortingType", lsqElmnt));
			core.lsqPower = getEnergyConfig(lsqElmnt);
			
			Element iTLBElmnt = (Element)(coreElmnt.getElementsByTagName("ITLB")).item(0);
			core.ITLBSize = Integer.parseInt(getImmediateString("Size", iTLBElmnt));
			core.ITLBLatency = Integer.parseInt(getImmediateString("Latency", iTLBElmnt));
			core.ITLBMissPenalty = Integer.parseInt(getImmediateString("MissPenalty", iTLBElmnt));
			core.ITLBPortType = setPortType(getImmediateString("PortType", iTLBElmnt));
			core.ITLBAccessPorts = Integer.parseInt(getImmediateString("AccessPorts", iTLBElmnt));
			core.ITLBPortOccupancy = Integer.parseInt(getImmediateString("PortOccupancy", iTLBElmnt));
			core.iTLBPower = getEnergyConfig(iTLBElmnt);
			
			Element dTLBElmnt = (Element)(coreElmnt.getElementsByTagName("DTLB")).item(0);
			core.DTLBSize = Integer.parseInt(getImmediateString("Size", dTLBElmnt));
			core.DTLBLatency = Integer.parseInt(getImmediateString("Latency", dTLBElmnt));
			core.DTLBMissPenalty = Integer.parseInt(getImmediateString("MissPenalty", dTLBElmnt));
			core.DTLBPortType = setPortType(getImmediateString("PortType", dTLBElmnt));
			core.DTLBAccessPorts = Integer.parseInt(getImmediateString("AccessPorts", dTLBElmnt));
			core.DTLBPortOccupancy = Integer.parseInt(getImmediateString("PortOccupancy", dTLBElmnt));
			core.dTLBPower = getEnergyConfig(dTLBElmnt);

			Element decodeElmnt = (Element)(coreElmnt.getElementsByTagName("Decode")).item(0);
			core.DecodeWidth = Integer.parseInt(getImmediateString("Width", decodeElmnt));
			core.decodePower = getEnergyConfig(decodeElmnt);
			
			Element instructionWindowElmnt = (Element)(coreElmnt.getElementsByTagName("InstructionWindow")).item(0);
			core.IssueWidth = Integer.parseInt(getImmediateString("IssueWidth", instructionWindowElmnt));			
			core.IWSize = Integer.parseInt(getImmediateString("IWSize", instructionWindowElmnt));
			core.iwPower = getEnergyConfig(instructionWindowElmnt);

			Element robElmnt = (Element)(coreElmnt.getElementsByTagName("ROB")).item(0);
			core.RetireWidth = Integer.parseInt(getImmediateString("RetireWidth", coreElmnt));
			core.ROBSize = Integer.parseInt(getImmediateString("ROBSize", coreElmnt));
			core.robPower = getEnergyConfig(robElmnt);
			
			Element resultsBroadcastBusElmnt = (Element)(coreElmnt.getElementsByTagName("ResultsBroadcastBus")).item(0);
			core.resultsBroadcastBusPower = getEnergyConfig(resultsBroadcastBusElmnt);
			
			Element renameElmnt = (Element)(coreElmnt.getElementsByTagName("Rename")).item(0);
			
			Element ratElmnt = (Element)(renameElmnt.getElementsByTagName("RAT")).item(0);
			core.intRATPower = getEnergyConfig((Element)ratElmnt.getElementsByTagName("Integer").item(0));
			core.floatRATPower = getEnergyConfig((Element)ratElmnt.getElementsByTagName("Float").item(0));
			
			Element freelistElmnt = (Element)(renameElmnt.getElementsByTagName("FreeList")).item(0);
			core.intFreeListPower = getEnergyConfig((Element)freelistElmnt.getElementsByTagName("Integer").item(0));
			core.floatFreeListPower = getEnergyConfig((Element)freelistElmnt.getElementsByTagName("Float").item(0));			
			
			Element registerFileElmnt = (Element)(coreElmnt.getElementsByTagName("RegisterFile")).item(0);
			
			Element integerRegisterFileElmnt = (Element)(registerFileElmnt.getElementsByTagName("Integer")).item(0);
			core.IntRegFileSize = Integer.parseInt(getImmediateString("IntRegFileSize", integerRegisterFileElmnt));
			core.IntArchRegNum = Integer.parseInt(getImmediateString("IntArchRegNum", integerRegisterFileElmnt));
			core.intRegFilePower = getEnergyConfig(integerRegisterFileElmnt);
			
			Element floatRegisterFileElmnt = (Element)(registerFileElmnt.getElementsByTagName("Float")).item(0);
			core.FloatRegFileSize = Integer.parseInt(getImmediateString("FloatRegFileSize", floatRegisterFileElmnt));
			core.FloatArchRegNum = Integer.parseInt(getImmediateString("FloatArchRegNum", floatRegisterFileElmnt));
			core.floatRegFilePower = getEnergyConfig(floatRegisterFileElmnt);
			
			Element intALUElmnt = (Element)(coreElmnt.getElementsByTagName("IntALU")).item(0);
			core.IntALUNum = Integer.parseInt(getImmediateString("IntALUNum", intALUElmnt));
			core.IntALULatency = Integer.parseInt(getImmediateString("IntALULatency", intALUElmnt));
			core.intALUPower = getEnergyConfig(intALUElmnt);

			Element floatALUElmnt = (Element)(coreElmnt.getElementsByTagName("FloatALU")).item(0);
			core.FloatALUNum = Integer.parseInt(getImmediateString("FloatALUNum", floatALUElmnt));
			core.FloatALULatency = Integer.parseInt(getImmediateString("FloatALULatency", floatALUElmnt));
			core.floatALUPower = getEnergyConfig(floatALUElmnt);
			
			Element complexALUElmnt = (Element)(coreElmnt.getElementsByTagName("ComplexALU")).item(0);
			core.IntMulNum = Integer.parseInt(getImmediateString("IntMulNum", complexALUElmnt));
			core.IntDivNum = Integer.parseInt(getImmediateString("IntDivNum", complexALUElmnt));
			core.FloatMulNum = Integer.parseInt(getImmediateString("FloatMulNum", complexALUElmnt));
			core.FloatDivNum = Integer.parseInt(getImmediateString("FloatDivNum", complexALUElmnt));
			core.IntMulLatency = Integer.parseInt(getImmediateString("IntMulLatency", complexALUElmnt));
			core.IntDivLatency = Integer.parseInt(getImmediateString("IntDivLatency", complexALUElmnt));
			core.FloatMulLatency = Integer.parseInt(getImmediateString("FloatMulLatency", complexALUElmnt));
			core.FloatDivLatency = Integer.parseInt(getImmediateString("FloatDivLatency", complexALUElmnt));
			core.complexALUPower = getEnergyConfig(complexALUElmnt);
						
			//set Branch Predictor Parameters
			core.branchPredictor = new BranchPredictorConfig();
			Element predictorElmnt = (Element)(coreElmnt.getElementsByTagName("BranchPredictor").item(0));
			Element BTBElmnt = (Element) predictorElmnt.getElementsByTagName("BTB").item(0);
			setBranchPredictorProperties(predictorElmnt, BTBElmnt, core.branchPredictor);
			core.BranchMispredPenalty = Integer.parseInt(getImmediateString("BranchMispredPenalty", predictorElmnt));
			core.bPredPower = getEnergyConfig(predictorElmnt);
			
			if(getImmediateString("TreeBarrier", coreElmnt).compareTo("true") == 0)
				core.TreeBarrier = true;
			else
				core.TreeBarrier = false;
			core.barrierLatency = Integer.parseInt(getImmediateString("BarrierLatency", coreElmnt));
			
			String tempStr = getImmediateString("BarrierUnit", coreElmnt);
			if (tempStr.equalsIgnoreCase("Central"))
				core.barrierUnit = 0;
			else if (tempStr.equalsIgnoreCase("Distributed"))
				core.barrierUnit = 1;
			else{
				System.err.println("Only Central and Distributed allowed as barrier unit");
				System.exit(0);
			}
			String interconnect = getImmediateString("Interconnect",systemElmnt); 
			if(interconnect.equalsIgnoreCase("Bus"))
			{
				SystemConfig.interconnect = Interconnect.Bus;
			}
			else if(interconnect.equalsIgnoreCase("Noc"))
			{
				SystemConfig.interconnect = Interconnect.Noc;
			}
			else
			{
				System.err.println("XML Configuration error : Invalid Interconnect Type");
				System.exit(1);
			}
			
			
			NodeList coreCacheList = coreElmnt.getElementsByTagName("Cache");
			if (coreCacheList.item(0) == null) {
				System.out.println("No core cache not found !!");
			} else {
				for(int coreCacheIndex=0; coreCacheIndex<coreCacheList.getLength(); coreCacheIndex++) {
					Element cacheNode = (Element)coreCacheList.item(coreCacheIndex);
					CacheConfig config = createCacheConfig(cacheNode);
					core.coreCacheList.add(config);
					
					// icache config
					if(SimulationConfig.collectInsnWorkingSetInfo &&
						config.firstLevel==true && 
						config.cacheDataType==CacheDataType.Instruction)
					{
						config.collectWorkingSetData = true;
						config.workingSetChunkSize = SimulationConfig.insnWorkingSetChunkSize; 
					}
					
					// l1cache config
					if(SimulationConfig.collectInsnWorkingSetInfo &&
						config.firstLevel==true && 
						config.cacheDataType==CacheDataType.Data)
					{
						config.collectWorkingSetData = true;
						config.workingSetChunkSize = SimulationConfig.dataWorkingSetChunkSize; 
					}
				}
			}
		}
		
		//Set Directory Parameters
		SystemConfig.directoryConfig = new CacheConfig();
		NodeList dirLst=systemElmnt.getElementsByTagName("Directory");
		Element dirElmnt = (Element) dirLst.item(0);
		setCacheProperties(dirElmnt, SystemConfig.directoryConfig);
		SystemConfig.directoryConfig.power = getCacheEnergyConfig(dirElmnt);
		
//		//Code for remaining Cache configurations
//		NodeList cacheLst = systemElmnt.getElementsByTagName("Cache");
//		for (int i = 0; i < cacheLst.getLength(); i++)
//		{
//			Element cacheElmnt = (Element) cacheLst.item(i);
//			String cacheName = cacheElmnt.getAttribute("name");
//
//			if (!(SystemConfig.declaredCaches.containsKey(cacheName)))	//If the identically named cache is not already present
//			{
//				CacheConfig newCacheConfigEntry = new CacheConfig();
////				newCacheConfigEntry.isFirstLevel = false;
//				newCacheConfigEntry.levelFromTop = Cache.CacheType.Lower;
//				String cacheType = cacheElmnt.getAttribute("type");
//				Element cacheTypeElmnt = searchLibraryForItem(cacheType);
//				setCacheProperties(cacheTypeElmnt, newCacheConfigEntry);
//				newCacheConfigEntry.nextLevel = cacheElmnt.getAttribute("nextLevel");
//				newCacheConfigEntry.operatingFreq = Long.parseLong(cacheElmnt.getAttribute("frequency"));
//				newCacheConfigEntry.power = getCachePowerConfig(cacheTypeElmnt);
//				SystemConfig.declaredCaches.put(cacheName, newCacheConfigEntry);
//			}
//		}
				
		
		
		
		//set NOC Parameters
		SystemConfig.nocConfig = new NocConfig();
		NodeList NocLst = systemElmnt.getElementsByTagName("NOC");
		Element nocElmnt = (Element) NocLst.item(0);
		SystemConfig.nocConfig.power = getEnergyConfig(nocElmnt);
		setNocProperties(nocElmnt, SystemConfig.nocConfig);
		
	}

	private static void setNocProperties(Element NocType, NocConfig nocConfig)
	{
		NodeList NocLst = NocType.getElementsByTagName("L2");
		Element NocElmnt = (Element) NocLst.item(0);
		setL2NocProperties(NocElmnt, nocConfig);
	}
	private static void setL2NocProperties(Element NocType, NocConfig nocConfig)
	{
		if(SystemConfig.interconnect==Interconnect.Noc)
		{
			String nocConfigFilename = getImmediateString("NocConfigFile", NocType);
			try 
			{
				File outputFile = new File(nocConfigFilename);
				if(!outputFile.exists()) 
				{
					System.err.println("XML Configuration error : NocConfigFile doesnot exist");
					System.exit(1);
				}
				
				BufferedReader readNocConfig = new BufferedReader(new FileReader(outputFile));
				String str;
				StringTokenizer st;
				str=readNocConfig.readLine();
				st = new StringTokenizer(str," ");
			
				nocConfig.nocElements = new NocElements(
						Integer.parseInt((String)st.nextElement()),
						Integer.parseInt((String)st.nextElement()));
				 
				nocConfig.numberOfBankColumns = nocConfig.nocElements.columns; 
				nocConfig.numberOfBankRows = nocConfig.nocElements.rows;
				
				for(int i=0;i<nocConfig.nocElements.rows;i++)
				{
					str=readNocConfig.readLine();
					st = new StringTokenizer(str," ");
					nocConfig.nocElements.nocElementsLocations.add(new Vector<String>());
					for(int j=0;j<nocConfig.nocElements.columns;j++)
					{
						nocConfig.nocElements.nocElementsLocations.get(i).add((String)st.nextElement());
						if(nocConfig.nocElements.nocElementsLocations.get(i).get(j).equals("1"))
							nocConfig.nocElements.noOfCores++;
						else if(nocConfig.nocElements.nocElementsLocations.get(i).get(j).equals("0"))
							nocConfig.nocElements.noOfCacheBanks++;
						else if(nocConfig.nocElements.nocElementsLocations.get(i).get(j).equals("D"))
							nocConfig.nocElements.noOfL1Directories++;
						else if(nocConfig.nocElements.nocElementsLocations.get(i).get(j).equals("M"))
							nocConfig.nocElements.noOfMemoryControllers++;
					}
				}
			}
			catch(Exception e)
			{
				System.err.println(e);
				System.exit(0);
			}
		}
		nocConfig.numberOfBuffers = Integer.parseInt(getImmediateString("NocNumberOfBuffers", NocType));
		nocConfig.portType = setPortType(getImmediateString("NocPortType", NocType));
		nocConfig.accessPorts = Integer.parseInt(getImmediateString("NocAccessPorts", NocType));
		nocConfig.portOccupancy = Integer.parseInt(getImmediateString("NocPortOccupancy", NocType));
		nocConfig.latency = Integer.parseInt(getImmediateString("NocLatency", NocType));
		nocConfig.operatingFreq = Integer.parseInt(getImmediateString("NocOperatingFreq", NocType));
		nocConfig.latencyBetweenBanks = Integer.parseInt(getImmediateString("NocLatencyBetweenBanks", NocType));
		
		String tempStr = getImmediateString("NucaMapping", NocType);
		if (tempStr.equalsIgnoreCase("S"))
			nocConfig.mapping = Mapping.SET_ASSOCIATIVE;
		else if (tempStr.equalsIgnoreCase("A"))
			nocConfig.mapping = Mapping.ADDRESS;
		else
		{
			System.err.println("XML Configuration error : Invalid value of 'NucaMapping' (please enter 'S'or 'A')");
			System.exit(1);
		}
		
		tempStr = getImmediateString("NocTopology", NocType);
		if(tempStr.equalsIgnoreCase("MESH"))
			nocConfig.topology = NOC.TOPOLOGY.MESH;
		else if(tempStr.equalsIgnoreCase("TORUS"))
			nocConfig.topology = NOC.TOPOLOGY.TORUS;
		else if(tempStr.equalsIgnoreCase("BUS"))
			nocConfig.topology = NOC.TOPOLOGY.BUS;
		else if(tempStr.equalsIgnoreCase("RING"))
			nocConfig.topology = NOC.TOPOLOGY.RING;
		else if(tempStr.equalsIgnoreCase("FATTREE"))
			nocConfig.topology = NOC.TOPOLOGY.FATTREE;
		else if(tempStr.equalsIgnoreCase("OMEGA"))
			nocConfig.topology = NOC.TOPOLOGY.OMEGA;
		else if(tempStr.equalsIgnoreCase("BUTTERFLY"))
			nocConfig.topology = NOC.TOPOLOGY.BUTTERFLY;
		
		tempStr = getImmediateString("NocRoutingAlgorithm", NocType);
		if(tempStr.equalsIgnoreCase("SIMPLE"))
			nocConfig.rAlgo = RoutingAlgo.ALGO.SIMPLE;
		else if(tempStr.equalsIgnoreCase("WESTFIRST"))
			nocConfig.rAlgo = RoutingAlgo.ALGO.WESTFIRST;
		else if(tempStr.equalsIgnoreCase("NORTHLAST"))
			nocConfig.rAlgo = RoutingAlgo.ALGO.NORTHLAST;
		else if(tempStr.equalsIgnoreCase("NEGATIVEFIRST"))
			nocConfig.rAlgo = RoutingAlgo.ALGO.NEGATIVEFIRST;
		else if(tempStr.equalsIgnoreCase("FATTREE"))
			nocConfig.rAlgo = RoutingAlgo.ALGO.FATTREE;
		else if(tempStr.equalsIgnoreCase("OMEGA"))
			nocConfig.rAlgo = RoutingAlgo.ALGO.OMEGA;
		else if(tempStr.equalsIgnoreCase("BUTTERFLY"))
			nocConfig.rAlgo = RoutingAlgo.ALGO.BUTTERFLY;

		tempStr = getImmediateString("NocSelScheme", NocType);
		if(tempStr.equalsIgnoreCase("STATIC"))
			nocConfig.selScheme = RoutingAlgo.SELSCHEME.STATIC;
		else
			nocConfig.selScheme = RoutingAlgo.SELSCHEME.ADAPTIVE;
		tempStr = getImmediateString("NocRouterArbiter", NocType);
		if(tempStr.equalsIgnoreCase("RR"))
			nocConfig.arbiterType = RoutingAlgo.ARBITER.RR_ARBITER;
		else if(tempStr.equalsIgnoreCase("MATRIX"))
			nocConfig.arbiterType = RoutingAlgo.ARBITER.MATRIX_ARBITER;
		else
			nocConfig.arbiterType = RoutingAlgo.ARBITER.QUEUE_ARBITER;
		nocConfig.technologyPoint = Integer.parseInt(getImmediateString("TechPoint", NocType));
		tempStr = getImmediateString("NocConnection", NocType);
		if(tempStr.equalsIgnoreCase("ELECTRICAL"))
			nocConfig.ConnType = CONNECTIONTYPE.ELECTRICAL;
		else
			nocConfig.ConnType = CONNECTIONTYPE.OPTICAL;	
	}
	private static void setCacheProperties(Element CacheType, CacheConfig cache)
	{
		String tempStr = getImmediateString("WriteMode", CacheType);
		if (tempStr.equalsIgnoreCase("WB"))
			cache.writePolicy = CacheConfig.WritePolicy.WRITE_BACK;
		else if (tempStr.equalsIgnoreCase("WT"))
			cache.writePolicy = CacheConfig.WritePolicy.WRITE_THROUGH;
		else
		{
			System.err.println("XML Configuration error : Invalid Write Mode (please enter WB for write-back or WT for write-through)");
			System.exit(1);
		}
		
		//System.out.println(cache.writeMode);
		
		cache.blockSize = Integer.parseInt(getImmediateString("BlockSize", CacheType));
		cache.assoc = Integer.parseInt(getImmediateString("Associativity", CacheType));
		
		if(isElementPresent("Size", CacheType)) {
			cache.size = Integer.parseInt(getImmediateString("Size", CacheType));
		} else {
			cache.size = 0;
		}
		
		if(isElementPresent("NumEntries", CacheType)) {
			cache.numEntries = Integer.parseInt(getImmediateString("NumEntries", CacheType));
			cache.size = cache.numEntries*cache.blockSize;
			System.out.println("Number of entries in directory : " + cache.numEntries);
			System.out.println("Directory size in terms of bytes (Considering one directory entry = one cache line ) : " + cache.size);
		} else {
			cache.numEntries = 0;
		}
		
		if(cache.size==0 && cache.numEntries==0) {
			misc.Error.showErrorAndExit("Invalid cache configuration : size=0 and numEntries=0 !!");
		}
		
		cache.latency = Integer.parseInt(getImmediateString("Latency", CacheType));
		cache.portType = setPortType(getImmediateString("PortType", CacheType));
		cache.accessPorts = Integer.parseInt(getImmediateString("AccessPorts", CacheType));
		cache.portOccupancy = Integer.parseInt(getImmediateString("PortOccupancy", CacheType));
		cache.multiportType = setMultiPortingType(getImmediateString("MultiPortingType", CacheType));
		cache.mshrSize = Integer.parseInt(getImmediateString("MSHRSize", CacheType));
				
		tempStr = getImmediateString("Coherence", CacheType);
		if (tempStr.equalsIgnoreCase("N"))
			cache.coherence = CoherenceType.None;
		else if (tempStr.equalsIgnoreCase("S"))
			cache.coherence = CoherenceType.Snoopy;
		else if (tempStr.equalsIgnoreCase("D"))
			cache.coherence = CoherenceType.Directory;
		else
		{
			System.err.println("XML Configuration error : Invalid value of 'Coherence' (please enter 'S', D' or 'N')");
			System.exit(1);
		}
		cache.numberOfBuses = Integer.parseInt(getImmediateString("NumBuses", CacheType));
		cache.busOccupancy = Integer.parseInt(getImmediateString("BusOccupancy", CacheType));
		
		tempStr = getImmediateString("Nuca", CacheType);
		
		if(cache.levelFromTop==memorysystem.Cache.CacheType.Lower)
		{
			if(SystemConfig.interconnect == Interconnect.Noc)
			{
				if (tempStr.equalsIgnoreCase("S"))
				{
					SimulationConfig.nucaType = NucaType.S_NUCA;
					cache.nucaType = NucaType.S_NUCA;
				}
				else if (tempStr.equalsIgnoreCase("D"))
				{
					SimulationConfig.nucaType = NucaType.D_NUCA;
					cache.nucaType = NucaType.D_NUCA;
				}
				else
				{
					System.err.println("XML Configuration error : Invalid value of 'Nuca' (please enter 'S', D' for NOC interconnect " +
							"or Choose Bus interconnect)");
					System.exit(1);
				}
			}
			
			else if(SystemConfig.interconnect == Interconnect.Bus)
			{
				SimulationConfig.nucaType = NucaType.NONE;
				cache.nucaType = NucaType.NONE;
			}
		}
		else
		{
			cache.nucaType = NucaType.NONE;
		}
		
		cache.cacheDataType = CacheDataType.valueOf(getImmediateString("CacheType", CacheType));
		
		cache.power = getCacheEnergyConfig(CacheType);
	}
	private static void setBranchPredictorProperties(Element predictorElmnt, Element BTBElmnt, BranchPredictorConfig branchPredictor){
		
		String tempStr = getImmediateString("Predictor_Mode", predictorElmnt);
		if(tempStr.equalsIgnoreCase("NoPredictor"))
			branchPredictor.predictorMode = BP.NoPredictor;
		else if(tempStr.equalsIgnoreCase("PerfectPredictor"))
			branchPredictor.predictorMode = BP.PerfectPredictor;
		else if(tempStr.equalsIgnoreCase("AlwaysTaken"))
			branchPredictor.predictorMode = BP.AlwaysTaken;
		else if(tempStr.equalsIgnoreCase("AlwaysNotTaken"))
			branchPredictor.predictorMode = BP.AlwaysNotTaken;
		else if(tempStr.equalsIgnoreCase("Tournament"))
			branchPredictor.predictorMode = BP.Tournament;
		else if(tempStr.equalsIgnoreCase("Bimodal"))
			branchPredictor.predictorMode = BP.Bimodal;
		else if(tempStr.equalsIgnoreCase("GAg"))
			branchPredictor.predictorMode = BP.GAg;
		else if(tempStr.equalsIgnoreCase("GAp"))
			branchPredictor.predictorMode = BP.GAp;
		else if(tempStr.equalsIgnoreCase("GShare"))
			branchPredictor.predictorMode = BP.GShare;
		else if(tempStr.equalsIgnoreCase("PAg"))
			branchPredictor.predictorMode = BP.PAg;
		else if(tempStr.equalsIgnoreCase("PAp"))
			branchPredictor.predictorMode = BP.PAp;
		
		branchPredictor.PCBits = Integer.parseInt(getImmediateString("PCBits", predictorElmnt));
		branchPredictor.BHRsize = Integer.parseInt(getImmediateString("BHRsize", predictorElmnt));
		branchPredictor.saturating_bits = Integer.parseInt(getImmediateString("SaturatingBits", predictorElmnt));
	}
	
	private static boolean setDirectoryCoherent(String immediateString) {
		if(immediateString==null)
			return false;
		if(immediateString.equalsIgnoreCase("T"))
			return true;
		else
			return false;
	}

	private static Element searchLibraryForItem(String tagName)	//Searches the <Library> section for a given tag name and returns it in Element form
	{															// Used mainly for cache types
		NodeList nodeLst = doc.getElementsByTagName("Library");
		Element libraryElmnt = (Element) nodeLst.item(0);
		NodeList libItemLst = libraryElmnt.getElementsByTagName(tagName);
		
		if (libItemLst.item(0) == null) //Item not found
		{
			System.err.println("XML Configuration error : Item type \"" + tagName + "\" not found in library section in the configuration file!!");
			System.exit(1);
		}
		
		if (libItemLst.item(1) != null) //Item found more than once
		{
			System.err.println("XML Configuration error : More than one definitions of item type \"" + tagName + "\" found in library section in the configuration file!!");
			System.exit(1);
		}
		
		Element resultElmnt = (Element) libItemLst.item(0);
		return resultElmnt;
	}
	
	private static String getImmediateString(String tagName, Element parent) // Get the immediate string value of a particular tag name under a particular parent tag
	{
		NodeList nodeLst = parent.getElementsByTagName(tagName);
		if (nodeLst.item(0) == null)
		{
			System.err.println("XML Configuration error : Item \"" + tagName + "\" not found inside the \"" + parent.getTagName() + "\" tag in the configuration file!!");
			System.exit(1);
		}
	    Element NodeElmnt = (Element) nodeLst.item(0);
	    NodeList resultNode = NodeElmnt.getChildNodes();
	    return ((Node) resultNode.item(0)).getNodeValue();
	}
	
	private static MultiPortingType setMultiPortingType(String inputStr)
	{
		MultiPortingType result = null;
		if (inputStr.equalsIgnoreCase("G"))
			result = MultiPortingType.GENUINE;
		else if (inputStr.equalsIgnoreCase("B"))
			result = MultiPortingType.BANKED;
		else
		{
			System.err.println("XML Configuration error : Invalid Multiporting type specified");
			System.exit(1);
		}
		return result;
	}
	
	private static PortType setPortType(String inputStr)
	{
		PortType result = null;
		if (inputStr.equalsIgnoreCase("UL"))
			result = PortType.Unlimited;
		else if (inputStr.equalsIgnoreCase("FCFS"))
			result = PortType.FirstComeFirstServe;
		else if (inputStr.equalsIgnoreCase("PR"))
			result = PortType.PriorityBased;
		else
		{
			System.err.println("XML Configuration error : Invalid Port Type type specified");
			System.exit(1);
		}
		return result;
	}
}
